#include "php.h"
#include "php_ini.h"
#include "php_main.h"
#include "php_globals.h"
#include "php_variables.h"

#include "zend.h"
#include "zend_hash.h"
#include "zend_modules.h"
#include "zend_interfaces.h"

#include "SAPI.h"

#include <stdio.h>
#include <errno.h>
#include <getopt.h>
#include <libgen.h>

#include "ext/standard/php_standard.h"

static void print_help(char *prog, int ret)
{
	printf("Usage: %s [OPTIONS]\n"
	       "\n"
	       "Options:\n"
	       "  -c, --config=FILE     specify php.ini\n"
	       "  -G, --general         show general configuration.\n"
	       "  -T, --credits         show credits.\n"
	       "  -C, --configuration   show configuration.\n"
	       "  -M, --modules         show additional modules.\n"
	       "  -E, --environment     show runtime environment.\n"
	       "  -V, --variables       show PHP variables.\n"
	       "  -L, --license         show PHP license.\n"
	       "  -h, --help            output a brief help message.\n"
	       "  -i, --info            option ignored.\n"
	       "\n", basename(prog));
	exit(ret);
}

struct option long_options[] = {
	{ "help",		no_argument,		0, 'h' },
	{ "general",		no_argument,		0, 'G' },
	{ "credits",		no_argument,		0, 'T' },
	{ "configuration",	no_argument,		0, 'C' },
	{ "modules",		no_argument,		0, 'M' },
	{ "environment",	no_argument,		0, 'E' },
	{ "variables",		no_argument,		0, 'V' },
	{ "license",		no_argument,		0, 'L' },
	{ "info",		no_argument,		0, 'i' },
	{ "config",		required_argument,	0, 'c' },
	{ 0, 0, 0, 0 }
};

static int php_phpinfo_startup(sapi_module_struct *sapi_module)
{
	if (php_module_startup(sapi_module, NULL, 0)==FAILURE) {
		return FAILURE;
	}
	return SUCCESS;
}

static int sapi_phpinfo_deactivate(TSRMLS_D)
{
	fflush(stdout);
	if(SG(request_info).argv0) {
		free(SG(request_info).argv0);
		SG(request_info).argv0 = NULL;
	}
	return SUCCESS;
}

static size_t sapi_phpinfo_single_write(const char *str, uint str_length)
{
	long ret;
	if ((ret = write(1, str, str_length)) <= 0)
		ret = 0;
	return ret;
}

static int sapi_phpinfo_ub_write(const char *str, uint str_length TSRMLS_DC)
{
	const char *ptr = str;
	uint remaining = str_length;
	size_t ret;

	while (remaining > 0)
	{
		ret = sapi_phpinfo_single_write(ptr, remaining);
		if (!ret) {
			php_handle_aborted_connection();
		}
		ptr += ret;
		remaining -= ret;
	}

	return str_length;
}

static void sapi_phpinfo_flush(void *server_context)
{
	/* Ignore EBADF here, it's caused by the fact that STDIN/STDOUT/STDERR streams
	 * are/could be closed before fflush() is called.
	 */
	fflush(stdout);
}

static int sapi_phpinfo_header_handler(sapi_header_struct *h, sapi_header_op_enum op, sapi_headers_struct *s TSRMLS_DC)
{
	return 0;
}

static int sapi_phpinfo_send_headers(sapi_headers_struct *sapi_headers TSRMLS_DC)
{
	/* We do nothing here, this function is needed to prevent that the fallback
	 * header handling is called. */
	return SAPI_HEADER_SENT_SUCCESSFULLY;
}

static void sapi_phpinfo_send_header(sapi_header_struct *sapi_header, void *server_context TSRMLS_DC)
{
}

static char* sapi_phpinfo_read_cookies(TSRMLS_D)
{
	return NULL;
}


static void sapi_phpinfo_log_message(char *message)
{
	fprintf(stderr, "%s\n", message);
}


/* {{{ sapi_module_struct phpinfo_sapi_module
 */
static sapi_module_struct phpinfo_sapi_module = {
	"phpini",			/* name */
	"PHP config parser",		/* pretty name */

	php_phpinfo_startup,		/* startup */
	php_module_shutdown_wrapper,	/* shutdown */

	NULL,				/* activate */
	sapi_phpinfo_deactivate,		/* deactivate */

	sapi_phpinfo_ub_write,		/* unbuffered write */
	sapi_phpinfo_flush,			/* flush */
	NULL,				/* get uid */
	NULL,				/* getenv */

	php_error,			/* error handler */

	sapi_phpinfo_header_handler,	/* header handler */
	sapi_phpinfo_send_headers,		/* send headers handler */
	sapi_phpinfo_send_header,		/* send header handler */

	NULL,				/* read POST data */
	sapi_phpinfo_read_cookies,		/* read Cookies */

	NULL,				/* register server variables */
	sapi_phpinfo_log_message,		/* Log message */
	NULL,				/* Get request time */
	NULL,				/* Child terminate */

	STANDARD_SAPI_MODULE_PROPERTIES
};

static void show_usage(char *prog, char *msg, int retval)
{
	if (msg != NULL)
		fprintf(stderr, "Error: %s\n", msg);

	fprintf(stderr, "Try `%s --help' for more information.\n", basename(prog));

	exit(retval);
}

int main(int argc, char *argv[])
{
	int c;
	int flag = 0;
	int exit_status = EXIT_SUCCESS;
	char *php_ini = NULL;

#ifdef ZTS
	void ***tsrm_ls;

	tsrm_startup(1, 1, 0, NULL);
	tsrm_ls = ts_resource(0);
#endif

	while ((c = getopt_long (argc, argv, "hGTCMEVLic:", long_options, NULL)) != -1) {
		switch (c) {
			case 'G': flag |= PHP_INFO_GENERAL;		break;
			case 'T': flag |= PHP_INFO_CREDITS;		break;
			case 'C': flag |= PHP_INFO_CONFIGURATION;	break;
			case 'M': flag |= PHP_INFO_MODULES;		break;
			case 'E': flag |= PHP_INFO_ENVIRONMENT;		break;
			case 'V': flag |= PHP_INFO_VARIABLES;		break;
			case 'L': flag |= PHP_INFO_LICENSE;		break;
			case 'i': break;
			case 'c':
				php_ini = optarg;
				break;
			case 'h':
				print_help(argv[0], EXIT_SUCCESS);
				break;
			default:
				show_usage(argv[0],
					"Insufficient arguments.",
					EXIT_FAILURE);
		}
	}

	if (php_ini == NULL)
		show_usage(argv[0], "php.ini required. Use -c option.", EXIT_FAILURE);

	if (access(php_ini, R_OK) == -1) {
		fprintf(stderr, "%s: %s", php_ini, strerror(errno));
		show_usage(argv[0], NULL, EXIT_FAILURE);
	}

	phpinfo_sapi_module.phpinfo_as_text = 1;
	phpinfo_sapi_module.executable_location = argv[0];
	phpinfo_sapi_module.php_ini_path_override = php_ini;

	sapi_startup(&phpinfo_sapi_module);

	/* startup after we get the above ini override se we get things right */
	if (phpinfo_sapi_module.startup(&phpinfo_sapi_module) == FAILURE) {
		fprintf(stderr, "Error!\n");
		exit(EXIT_FAILURE);
	}

	CG(in_compilation) = 0;
	EG(uninitialized_zval_ptr) = NULL;

	if (php_request_startup(TSRMLS_C) != FAILURE) {
		php_print_info((flag == 0 ? PHP_INFO_ALL : flag) TSRMLS_CC);
		php_end_ob_buffers(1 TSRMLS_CC);
	}
	else {
		fprintf(stderr, "php_request_startup: Error!\n");
		exit(EXIT_FAILURE);
	}

	php_request_shutdown((void *) 0);

	if (exit_status == EXIT_SUCCESS)
		exit_status = EG(exit_status);

	if (phpinfo_sapi_module.ini_entries)
		free(phpinfo_sapi_module.ini_entries);

	sapi_shutdown();
#ifdef ZTS
	tsrm_shutdown();
#endif

	exit(exit_status);
}
